package ch.unige.solidify.util;

import java.io.IOException;
import java.net.URI;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.web.client.RestTemplateBuilder;
import org.springframework.boot.web.client.RestTemplateCustomizer;
import org.springframework.core.io.buffer.DataBuffer;
import org.springframework.core.io.buffer.DataBufferUtils;
import org.springframework.http.MediaType;
import org.springframework.http.client.ClientHttpRequestInterceptor;
import org.springframework.http.client.ClientHttpResponse;
import org.springframework.http.converter.HttpMessageConverter;
import org.springframework.http.converter.StringHttpMessageConverter;
import org.springframework.stereotype.Component;
import org.springframework.web.client.DefaultResponseErrorHandler;
import org.springframework.web.client.RestTemplate;
import org.springframework.web.reactive.function.client.WebClient;

import reactor.core.publisher.Flux;

import ch.unige.solidify.interceptor.BearerAuthorizationInterceptor;
import ch.unige.solidify.interceptor.HeaderRequestInterceptor;
import ch.unige.solidify.interceptor.PlusEncoderInterceptor;
import ch.unige.solidify.model.sso.SsoUser;
import ch.unige.solidify.security.TrustedTokenStorage;

@Component
public class RestClientTool {
  private static class UTF8MessageConverterCustomizer implements RestTemplateCustomizer {
    @Override
    public void customize(RestTemplate restTemplate) {
      final List<HttpMessageConverter<?>> converters = restTemplate.getMessageConverters();
      converters.removeIf(httpMessageConverter -> httpMessageConverter instanceof StringHttpMessageConverter);
      converters.add(0, new StringHttpMessageConverter(StandardCharsets.UTF_8));
    }
  }

  private static final Logger log = LoggerFactory.getLogger(RestClientTool.class);

  @Value("${solidify.oauth2.accesstoken:#{null}}")
  private String accessToken;

  private final TrustedTokenStorage trustedTokenStorage;

  public RestClientTool(TrustedTokenStorage trustedTokenStorage) {
    this.trustedTokenStorage = trustedTokenStorage;
    if (this.accessToken == null) {
      this.accessToken = System.getProperty("solidify.oauth2.accesstoken");
    }
  }

  public void downloadContent(URI uri, Path path) {
    final WebClient webClient = WebClient.create(uri.toString());
    String token = null;
    if (!StringTool.isNullOrEmpty(this.accessToken)) {
      token = this.accessToken;
    } else if (this.trustedTokenStorage.getTrustedToken() != null) {
      token = this.trustedTokenStorage.getTrustedToken();
    }
    Flux<DataBuffer> dataStream = webClient.get()
             .accept(MediaType.APPLICATION_OCTET_STREAM)
             .header("Authorization", "Bearer " + token)
             .retrieve()
             .bodyToFlux(DataBuffer.class);
    DataBufferUtils.write(dataStream, path).block();
  }

  public RestTemplate getClient(String url) {
    return this.getRestTemplateBuilder(url).build();
  }

  public RestTemplate getClientWithPlusEncoderInterceptor(String url) {
    return this.getRestTemplateBuilder(url).additionalInterceptors(new PlusEncoderInterceptor()).build();
  }

  private RestTemplateBuilder getRestTemplateBuilder(String urlString) {
    RestTemplateBuilder restTemplateBuilder = new RestTemplateBuilder();
    if (!StringTool.isNullOrEmpty(this.accessToken)) {
      // OAuth2 without end user: shared AccesToken
      restTemplateBuilder = restTemplateBuilder.rootUri(urlString)
              .additionalInterceptors(new BearerAuthorizationInterceptor(this.accessToken));
    } else if (this.trustedTokenStorage.getTrustedToken() != null) {
      // DLCM : Access token of trusted user
      restTemplateBuilder = restTemplateBuilder.rootUri(urlString).additionalInterceptors(
              new BearerAuthorizationInterceptor(this.trustedTokenStorage.getTrustedToken()));
    } else if (UserTool.getUserPrincipal() instanceof SsoUser) {
      // Rock Solid : AccesToken, usercontext, useruniqueID of end user for OIDC Autho2 protected MS
      final SsoUser user = (SsoUser) UserTool.getUserPrincipal();
      final List<ClientHttpRequestInterceptor> intercepters = new ArrayList<>();
      if (user.getUniqueId() != null) {
        intercepters.add(new HeaderRequestInterceptor(SsoUser.CONNECTED_USER_UNIQUE_ID,
                user.getUniqueId()));
      }
      if (user.getUserContext() != null) {
        intercepters.add(new HeaderRequestInterceptor(SsoUser.CONNECTED_USER_CONTEXT_SOURCE_NAME,
                user.getUserContext()));
      }
      if (UserTool.getUserPrincipal().getAccessToken() != null) {
        intercepters.add(new BearerAuthorizationInterceptor(
                UserTool.getUserPrincipal().getAccessToken()));
      }
      restTemplateBuilder = restTemplateBuilder.rootUri(urlString)
              .additionalInterceptors(intercepters);
    } else {
      // No Auth
      restTemplateBuilder = restTemplateBuilder.rootUri(urlString);
    }

    // Use UTF-8 message converter
    restTemplateBuilder = restTemplateBuilder.additionalCustomizers(new UTF8MessageConverterCustomizer());

    // Error handler
    restTemplateBuilder.errorHandler(new DefaultResponseErrorHandler() {
      @Override
      public void handleError(ClientHttpResponse response) throws IOException {
        log.error("Error {} on {}", response.getStatusText(), RestClientTool.this.getUrl(new URL(urlString)));
        super.handleError(response);
      }
    });
    return restTemplateBuilder;
  }

  private String getUrl(URL url) {
    String rootUrl = url.getProtocol() + "://" + url.getHost();
    if (url.getPort() > 0) {
      rootUrl += ":" + url.getPort();
    }
    return rootUrl + url.getPath();
  }

}
